/*
 *
 * Copyright (c) 2000 Scott Oaks and Henry Wong. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and
 * without fee is hereby granted.
 *
 * This sample source code is provided for example only,
 * on an unsupported, as-is basis. 
 *
 * AUTHOR MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. AUTHOR SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 *
 * THIS SOFTWARE IS NOT DESIGNED OR INTENDED FOR USE OR RESALE AS ON-LINE
 * CONTROL EQUIPMENT IN HAZARDOUS ENVIRONMENTS REQUIRING FAIL-SAFE
 * PERFORMANCE, SUCH AS IN THE OPERATION OF NUCLEAR FACILITIES, AIRCRAFT
 * NAVIGATION OR COMMUNICATION SYSTEMS, AIR TRAFFIC CONTROL, DIRECT LIFE
 * SUPPORT MACHINES, OR WEAPONS SYSTEMS, IN WHICH THE FAILURE OF THE
 * SOFTWARE COULD LEAD DIRECTLY TO DEATH, PERSONAL INJURY, OR SEVERE
 * PHYSICAL OR ENVIRONMENTAL DAMAGE ("HIGH RISK ACTIVITIES").  AUTHOR
 * SPECIFICALLY DISCLAIMS ANY EXPRESS OR IMPLIED WARRANTY OF FITNESS FOR
 * HIGH RISK ACTIVITIES.
 */


import java.io.*;
import java.rmi.*;
import java.rmi.server.*;
import java.util.*;
import net.jini.core.lease.*;
import com.sun.jini.lease.landlord.*;

public class ServerLandlord extends UnicastRemoteObject
                            implements Landlord {
    protected Hashtable expiration = new Hashtable();
                            // cookie -> expiration
    protected Hashtable leases = new Hashtable();
                            // cookie -> lease
    protected Hashtable session = new Hashtable();   
                            // cookie -> sessiondata
    protected long MAXLEASETIME = 10*60*1000;         // 10 Minutes
    protected LandlordLease.Factory factory =
                                    new LandlordLease.Factory();

    public ServerLandlord() throws RemoteException {
    }

    public long renew(Object cookie, long duration)
            throws LeaseDeniedException,
                   UnknownLeaseException, RemoteException {

        // Time when the client first attempted renewal
        long now = System.currentTimeMillis();

        // Reset the duration to the maximum that were willing to give
        if (duration == Lease.ANY ||
            duration == Lease.FOREVER || 
            duration > MAXLEASETIME)
            duration = MAXLEASETIME;

        synchronized (this) {
            Long expire = (Long) expiration.get(cookie);
            if (expire == null)
                throw new UnknownLeaseException();
            // If the lease has already expired, dont renew it
            if (expire.longValue() < now) {
                expiration.remove(cookie);
                leases.remove(cookie);
                session.remove(cookie);
                throw new UnknownLeaseException();
            }
            expiration.remove(cookie);
            expiration.put(cookie, new Long(now + duration));
        }
        return duration;
    }

    public Landlord.RenewResults renewAll(Object[] cookie,
                            long[] duration) throws RemoteException {
        long granted[] = new long[cookie.length];
        Vector denied = new Vector();

        // Try to renew each lease. When one fails, place it in the
        // array of exceptions to be held in the RenewResults object
        for (int i = 0; i < cookie.length; i++) {
            try {
                granted[i] = renew(cookie[i], duration[i]);
            } catch (LeaseException lex) {
                granted[i] = -1;
                denied.add(lex);
            }
        }
        return new Landlord.RenewResults(granted,
                   denied.isEmpty() ? null :
                                      (Exception[]) denied.toArray());
    }

    public void cancel(Object cookie) 
                       throws UnknownLeaseException, RemoteException {
        synchronized (this) {
            if (expiration.get(cookie) == null) 
                throw new UnknownLeaseException();
            expiration.remove(cookie);
            leases.remove(cookie);
            session.remove(cookie);
        }
    }

    public void cancelAll(Object[] cookie) 
            throws LeaseMapException, RemoteException {
        Map map = null;
        // Cancel all the leases. If one or more has an exception,
        // then continue but save the exception in a map to be
        // thrown back to the user.
        for (int i = 0; i < cookie.length; i++) {
            try {
                cancel(cookie[i]);
            } catch (LeaseException ex) {
                if (map == null)
                    map = new HashMap();
                map.put(cookie[i], ex);
            }
        }
        if (map != null)
            throw new LeaseMapException("Can't cancel all leases", map);
    }

    // The following methods are not part of the Landlord interface
    // and can be called only by directly using this object.
    static int token = 0;            // Unique ID for cookie
    public Lease newLease(Object sessionData, long duration) {
        long now = System.currentTimeMillis();  // Time when requested

        Integer cookie;                // Get unique cookie
        synchronized(ServerLandlord.class) {
            cookie = new Integer(token++);
        }

        expireLeases();                // Make room for new Leases
        if (duration == Lease.ANY ||
            duration == Lease.FOREVER ||
            duration > MAXLEASETIME)
            duration = MAXLEASETIME;

        synchronized(this) {
            Lease l = factory.newLease(cookie, this, duration);
            expiration.put(cookie, new Long(now + duration));
            leases.put(cookie, l);
            if (sessionData != null) {
                session.put(cookie, sessionData);
            }
            return l;
        }
    }

    public void expireLeases() {
        long now = System.currentTimeMillis();
        now = now - 60 * 1000;          // Be conservative by one minute

        synchronized (this) {
            Vector deleteList = new Vector();
            for (Enumeration e = expiration.keys();
                             e.hasMoreElements();) {
                Object cookie = e.nextElement();
                Long expire = (Long) expiration.get(cookie);
                if (expire.longValue() < now) {
                    deleteList.addElement(cookie);
                }
            }
            for (Enumeration e = deleteList.elements();
                             e.hasMoreElements();) {
                Object cookie = e.nextElement();
                expiration.remove(cookie);
                leases.remove(cookie);
                session.remove(cookie);
            }
        }
    }

    // The cookie is an *internal* reference for the LandlordLease.
    // No one outside the lease can access it.
    public Object getSessionData(Lease lease) {
        Object sessiondata = null;
        if (lease instanceof LandlordLease) {
            expireLeases();        // Make sure the lease is still valid
            synchronized (this) {
                for (Enumeration e = leases.keys();
                                 e.hasMoreElements();) {
                    Object cookie = e.nextElement();
                    if (lease.equals(leases.get(cookie))) {
                        sessiondata = session.get(cookie);
                        break;
                    }
                }
            }
        }
        return sessiondata;
    }
}
